@@ -26,21 +26,16 @@
 
 import sys
 import inspect
+import itertools
+import weakref
 from ..overrides import override
-from ..importer import modules
+from ..module import get_introspection_module
 
 from gi.repository import GLib
 
 
-if sys.version_info >= (3, 0):
-    _basestring = str
-    _callable = lambda c: hasattr(c, '__call__')
-    long = int
-else:
-    _basestring = basestring
-    _callable = callable
+Gst = get_introspection_module('Gst')
 
-Gst = modules['Gst']._introspection_module
 __all__ = []
 
 if Gst._version == '0.10':
@@ -49,11 +44,27 @@ if Gst._version == '0.10':
 was not designed for use with introspection some of the \
 interfaces and API will fail.  As such this is not supported \
 by the GStreamer development team and we encourage you to \
-port your app to Gst 1 or greater. gst-python is the recomended \
+port your app to Gst 1 or greater. gst-python is the recommended \
 python module to use with Gst 0.10"
 
     warnings.warn(warn_msg, RuntimeWarning)
 
+
+class Element(Gst.Element):
+    @staticmethod
+    def link_many(*args):
+        '''
+        @raises: Gst.LinkError
+        '''
+        for pair in pairwise(args):
+            if not pair[0].link(pair[1]):
+                raise LinkError(
+                    'Failed to link {} and {}'.format(pair[0], pair[1]))
+
+Element = override(Element)
+__all__.append('Element')
+
+
 class Bin(Gst.Bin):
     def __init__(self, name=None):
         Gst.Bin.__init__(self, name=name)
@@ -63,6 +74,18 @@ class Bin(Gst.Bin):
             if not Gst.Bin.add(self, arg):
                 raise AddError(arg)
 
+    def make_and_add(self, factory_name, instance_name=None):
+        '''
+        @raises: Gst.AddError
+        '''
+        elem = Gst.ElementFactory.make(factory_name, instance_name)
+        if not elem:
+            raise AddError(
+                'No such element: {}'.format(factory_name))
+        self.add(elem)
+        return elem
+
+
 Bin = override(Bin)
 __all__.append('Bin')
 
@@ -109,59 +132,65 @@ class Caps(Gst.Caps):
 Caps = override(Caps)
 __all__.append('Caps')
 
-class Pad(Gst.Pad):
-    def __init__(self, *args, **kwargs):
-        self._real_chain_func = None
-        self._real_event_func = None
-        self._real_query_func = None
-        super(Gst.Pad, self).__init__(*args, **kwargs)
+class PadFunc:
+    def __init__(self, func):
+        self.func = func
 
-    def _chain_override(self, pad, parent, buf):
-        return self._real_chain_func(pad, buf)
+    def __call__(self, pad, parent, obj):
+        if isinstance(self.func, weakref.WeakMethod):
+            func = self.func()
+        else:
+            func = self.func
 
-    def _event_override(self, pad, parent, event):
-        return self._real_event_func(pad, event)
-
-    def _query_override(self, pad, parent, query):
-        query.mini_object.refcount -= 1
         try:
-            res = self._real_query_func(pad, query)
+            res = func(pad, obj)
         except TypeError:
             try:
-                res = self._real_query_func(pad, parent, query)
+                res = func(pad, parent, obj)
             except TypeError:
-                raise TypeError("Invalid query method %s, 2 or 3 arguments required"
-                                % self._real_query_func)
-        query.mini_object.refcount += 1
+                raise TypeError("Invalid method %s, 2 or 3 arguments required"
+                                % func)
 
         return res
 
+class Pad(Gst.Pad):
+    def __init__(self, *args, **kwargs):
+        super(Gst.Pad, self).__init__(*args, **kwargs)
+
     def set_chain_function(self, func):
-        self._real_chain_func = func
-        self.set_chain_function_full(self._chain_override, None)
+        self.set_chain_function_full(PadFunc(func), None)
 
     def set_event_function(self, func):
-        self._real_event_func = func
-        self.set_event_function_full(self._event_override, None)
+        self.set_event_function_full(PadFunc(func), None)
 
     def set_query_function(self, func):
-        self._real_query_func = func
-        self.set_query_function_full(self._chain_override, None)
-
-    def set_query_function_full(self, func, udata):
-        self._real_query_func = func
-        self._real_set_query_function_full(self._query_override, None)
+        self.set_query_function_full(PadFunc(func), None)
 
     def query_caps(self, filter=None):
         return Gst.Pad.query_caps(self, filter)
 
+    def set_caps(self, caps):
+        if not isinstance(caps, Gst.Caps):
+            raise TypeError("%s is not a Gst.Caps." % (type(caps)))
+
+        if not caps.is_fixed():
+            return False
+
+        event = Gst.Event.new_caps(caps)
+
+        if self.direction == Gst.PadDirection.SRC:
+            res = self.push_event(event)
+        else:
+            res = self.send_event(event)
+
+        return res
+
     def link(self, pad):
         ret = Gst.Pad.link(self, pad)
         if ret != Gst.PadLinkReturn.OK:
             raise LinkError(ret)
         return ret
 
-Pad._real_set_query_function_full = Gst.Pad.set_query_function_full
 Pad = override(Pad)
 __all__.append('Pad')
 
@@ -196,6 +225,11 @@ class LinkError(Exception):
     pass
 __all__.append('LinkError')
 
+class MapError(Exception):
+    pass
+__all__.append('MapError')
+
+
 class Iterator(Gst.Iterator):
     def __iter__(self):
         while True:
@@ -444,10 +478,10 @@ class Int64Range(Gst.Int64Range):
 
 class Bitmask(Gst.Bitmask):
     def __init__(self, v):
-        if not isinstance(v, long) and not isinstance(v, int):
-            raise TypeError("%s is not an int or long." % (type(v)))
+        if not isinstance(v, int):
+            raise TypeError("%s is not an int." % (type(v)))
 
-        self.v = long(v)
+        self.v = int(v)
 
     def __str__(self):
         return hex(self.v)
@@ -554,6 +588,80 @@ class ValueList(Gst.ValueList):
 ValueList = override(ValueList)
 __all__.append('ValueList')
 
+# From https://docs.python.org/3/library/itertools.html
+
+
+def pairwise(iterable):
+    a, b = itertools.tee(iterable)
+    next(b, None)
+    return zip(a, b)
+
+class MapInfo:
+    def __init__(self):
+        self.memory = None
+        self.flags = Gst.MapFlags(0)
+        self.size = 0
+        self.maxsize = 0
+        self.data = None
+        self.user_data = None
+        self.__parent__ = None
+
+    def __iter__(self):
+        # Make it behave like a tuple similar to the PyGObject generated API for
+        # the `Gst.Buffer.map()` and friends.
+        for i in (self.__parent__ is not None, self):
+            yield i
+
+    def __enter__(self):
+        if not self.__parent__:
+            raise MapError('MappingError', 'Mapping was not successful')
+
+        return self
+
+    def __exit__(self, type, value, tb):
+        if not self.__parent__.unmap(self):
+            raise MapError('MappingError', 'Unmapping was not successful')
+
+__all__.append("MapInfo")
+
+class Buffer(Gst.Buffer):
+
+    def map_range(self, idx, length, flags):
+        mapinfo = MapInfo()
+        if (_gi_gst.buffer_override_map_range(self, mapinfo, idx, length, int(flags))):
+            mapinfo.__parent__ = self
+
+        return mapinfo
+
+    def map(self, flags):
+        mapinfo = MapInfo()
+        if _gi_gst.buffer_override_map(self, mapinfo, int(flags)):
+            mapinfo.__parent__ = self
+
+        return mapinfo
+
+    def unmap(self, mapinfo):
+        mapinfo.__parent__ = None
+        return _gi_gst.buffer_override_unmap(self, mapinfo)
+
+Buffer = override(Buffer)
+__all__.append('Buffer')
+
+class Memory(Gst.Memory):
+
+    def map(self, flags):
+        mapinfo = MapInfo()
+        if (_gi_gst.memory_override_map(self, mapinfo, int(flags))):
+            mapinfo.__parent__ = self
+
+        return mapinfo
+
+    def unmap(self, mapinfo):
+        mapinfo.__parent__ = None
+        return _gi_gst.memory_override_unmap(self, mapinfo)
+
+Memory = override(Memory)
+__all__.append('Memory')
 
 def TIME_ARGS(time):
     if time == Gst.CLOCK_TIME_NONE:
@@ -565,7 +673,7 @@ def TIME_ARGS(time):
                                   time % Gst.SECOND)
 __all__.append('TIME_ARGS')
 
-from . import _gi_gst
+from gi.overrides import _gi_gst
 _gi_gst
 
 # maybe more python and less C some day if core turns a bit more introspection
@@ -609,7 +717,7 @@ def init_pygst():
 
 def deinit_pygst():
     for fname, func in real_functions:
-        if fname not in ["init", "init_check", "deinit"]:
+        if fname not in ["init", "init_check", "deinit", "is_initialized"]:
             setattr(Gst, fname, fake_method)
     for cname_class, methods in class_methods:
         for mname, method in methods:
