@@ -32,7 +32,6 @@
 
 #include <locale.h>
 
-#if PY_MAJOR_VERSION >= 3
 #define PYGLIB_MODULE_START(symbol, modname)	        \
     static struct PyModuleDef _##symbol##module = {     \
     PyModuleDef_HEAD_INIT,                              \
@@ -51,22 +50,13 @@ PyMODINIT_FUNC PyInit_##symbol(void)                    \
     PyObject *module;                                   \
     module = PyModule_Create(&_##symbol##module);
 #define PYGLIB_MODULE_END return module; }
-#else
-#define PYGLIB_MODULE_START(symbol, modname)            \
-DL_EXPORT(void) init##symbol(void);         \
-DL_EXPORT(void) init##symbol(void)          \
-{                                                       \
-    PyObject *module;                                   \
-    module = Py_InitModule(modname, symbol##_functions);
-#define PYGLIB_MODULE_END }
-#endif
 
 GST_DEBUG_CATEGORY_STATIC (python_debug);
 GST_DEBUG_CATEGORY_STATIC (pygst_debug);
 #define GST_CAT_DEFAULT pygst_debug
 
 static PyObject *
-gi_gst_get_type (gchar * type_name)
+gi_gst_get_type (const gchar * type_name)
 {
   PyObject *module, *dict;
 
@@ -81,7 +71,7 @@ gi_gst_get_type (gchar * type_name)
   dict = PyModule_GetDict (module);
   Py_DECREF (module);
 
-  /* For some reson we need this intermediary step */
+  /* For some reason we need this intermediary step */
   module = PyMapping_GetItemString (dict, "_overrides_module");
   if (module == NULL) {
     PyErr_SetString (PyExc_KeyError,
@@ -493,7 +483,7 @@ fail:
   return -1;
 }
 
-void
+static void
 gi_gst_register_types (PyObject * d)
 {
   pyg_register_gtype_custom (GST_TYPE_FRACTION,
@@ -654,7 +644,6 @@ pygst_debug_log (PyObject * pyobject, PyObject * string, GstDebugLevel level,
   }
 
   frame = PyEval_GetFrame ();
-#if PY_MAJOR_VERSION >= 3
   {
     PyObject *utf8;
     const gchar *utf8_str;
@@ -671,11 +660,6 @@ pygst_debug_log (PyObject * pyobject, PyObject * string, GstDebugLevel level,
     filename = g_strdup (utf8_str);
     Py_DECREF (utf8);
   }
-#else
-  function = g_strdup (PyString_AsString (frame->f_code->co_name));
-  filename =
-      g_path_get_basename (PyString_AsString (frame->f_code->co_filename));
-#endif
   lineno = PyCode_Addr2Line (frame->f_code, frame->f_lasti);
   /* gst_debug_log : category, level, file, function, line, object, format, va_list */
   if (isgstobject)
@@ -739,6 +723,298 @@ _wrap_gst_memdump (PyObject * whatever, PyObject * string)
   return pygst_debug_log (whatever, string, GST_LEVEL_MEMDUMP, FALSE);
 }
 
+static PyObject *
+_remap (GstMapInfo * mapinfo, PyObject * py_mapinfo)
+{
+  PyObject *success = NULL;
+  PyObject *py_cmapinfo = NULL;
+  PyObject *py_mview = NULL;
+  PyObject *py_memory = NULL;
+  PyObject *py_flags = NULL;
+  PyObject *py_size = NULL;
+  PyObject *py_maxsize = NULL;
+
+  /* Fill and encapsulating the mapinfo pointer */
+  py_cmapinfo = PyCapsule_New (mapinfo, "__cmapinfo", NULL);
+  if (!py_cmapinfo
+      || PyObject_SetAttrString (py_mapinfo, "__cmapinfo", py_cmapinfo))
+    goto err;
+
+  /* Fill and create memoryview with compatible flags */
+  int flags;
+  flags = (mapinfo->flags & GST_MAP_WRITE) ? PyBUF_WRITE : PyBUF_READ;
+  py_mview =
+      PyMemoryView_FromMemory ((char *) mapinfo->data, mapinfo->size, flags);
+  if (!py_mview || PyObject_SetAttrString (py_mapinfo, "data", py_mview))
+    goto err;
+
+  /* Fill and box GstMemory into a Gst.Memory */
+  py_memory = pyg_boxed_new (_gst_memory_type, mapinfo->memory, FALSE, FALSE);
+  if (!py_memory || PyObject_SetAttrString (py_mapinfo, "memory", py_memory))
+    goto err;
+
+  /* Fill out Gst.MapInfo with values corresponding to GstMapInfo */
+  py_flags = Py_BuildValue ("i", mapinfo->flags);
+  if (!py_flags || PyObject_SetAttrString (py_mapinfo, "flags", py_flags))
+    goto err;
+
+  py_size = Py_BuildValue ("i", mapinfo->size);
+  if (!py_size || PyObject_SetAttrString (py_mapinfo, "size", py_size))
+    goto err;
+
+  py_maxsize = Py_BuildValue ("i", mapinfo->maxsize);
+  if (!py_maxsize || PyObject_SetAttrString (py_mapinfo, "maxsize", py_maxsize))
+    goto err;
+
+  Py_INCREF (Py_True);
+  success = Py_True;
+  goto end;
+
+err:
+  GST_ERROR ("Could not map the Gst.MapInfo PyObject with GstMapInfo");
+  if (py_mview)
+    PyObject_CallMethod (py_mview, "release", NULL);
+
+end:
+  Py_XDECREF (py_cmapinfo);
+  Py_XDECREF (py_mview);
+  Py_XDECREF (py_memory);
+  Py_XDECREF (py_flags);
+  Py_XDECREF (py_size);
+  Py_XDECREF (py_maxsize);
+  return success;
+}
+
+static PyObject *
+_unmap (GstMapInfo ** mapinfo, PyObject * py_mapinfo)
+{
+  PyObject *py_cmapinfo = NULL, *py_mview = NULL, *success = NULL;
+
+  if (!PyObject_HasAttrString (py_mapinfo, "__cmapinfo"))
+    goto done;
+
+  /* Extract attributes from Gst.MapInfo */
+  py_mview = PyObject_GetAttrString (py_mapinfo, "data");
+  if (!py_mview)
+    goto err;
+
+  /* Call the memoryview.release() Python method, there is no C API */
+  if (!PyObject_CallMethod (py_mview, "release", NULL))
+    goto err;
+
+  py_cmapinfo = PyObject_GetAttrString (py_mapinfo, "__cmapinfo");
+  if (!py_cmapinfo)
+    goto err;
+
+  /* Reconstruct GstMapInfo from Gst.MapInfo contents */
+  *mapinfo = PyCapsule_GetPointer (py_cmapinfo, "__cmapinfo");
+  if (!*mapinfo)
+    goto err;
+
+  if (PyObject_DelAttrString (py_mapinfo, "__cmapinfo") == -1)
+    goto err;
+
+done:
+  Py_INCREF (Py_True);
+  success = Py_True;
+  goto end;
+
+err:
+  GST_ERROR ("Could not unmap the GstMapInfo from Gst.MapInfo PyObject");
+  Py_INCREF (Py_False);
+  success = Py_False;
+
+end:
+  Py_XDECREF (py_mview);
+  Py_XDECREF (py_cmapinfo);
+  return success;
+}
+
+static PyObject *
+_gst_memory_override_map (PyObject * self, PyObject * args)
+{
+  PyTypeObject *gst_memory_type;
+  PyObject *py_memory, *py_mapinfo, *success;
+  int flags;
+  GstMemory *memory;
+  GstMapInfo *mapinfo;
+  _Bool ok;
+
+  /* Look up Gst.memory, Gst.MapInfo, and Gst.MapFlags parameters */
+  gst_memory_type = pygobject_lookup_class (_gst_memory_type);
+  if (!PyArg_ParseTuple (args, "O!Oi", gst_memory_type, &py_memory,
+          &py_mapinfo, &flags))
+    return NULL;
+
+  /* Since Python does only support r/o or r/w it has to be changed to either */
+  flags = (flags & GST_MAP_WRITE) ? GST_MAP_READWRITE : GST_MAP_READ;
+
+  /* Extract GstMemory from Gst.Memory parameter */
+  memory = GST_MEMORY_CAST (pygobject_get (py_memory));
+
+  /* Map the memory, fill out GstMapInfo */
+  mapinfo = g_new0 (GstMapInfo, 1);
+  ok = gst_memory_map (memory, mapinfo, flags);
+  if (!ok) {
+    g_free (mapinfo);
+    goto err;
+  }
+
+  success = _remap (mapinfo, py_mapinfo);
+  if (!success) {
+    gst_memory_unmap (memory, mapinfo);
+    g_free (mapinfo);
+  }
+  return success;
+
+err:
+  Py_INCREF (Py_False);
+  return Py_False;
+}
+
+static PyObject *
+_gst_memory_override_unmap (PyObject * self, PyObject * args)
+{
+  PyTypeObject *gst_memory_type;
+  PyObject *py_memory, *py_mapinfo, *success;
+  GstMemory *memory;
+  GstMapInfo *mapinfo = NULL;
+
+  /* Look up Gst.Buffer and Gst.Mapinfo parameters */
+  gst_memory_type = pygobject_lookup_class (_gst_memory_type);
+  if (!PyArg_ParseTuple (args, "O!O", gst_memory_type, &py_memory, &py_mapinfo)) {
+    PyErr_BadArgument ();
+    return NULL;
+  }
+
+  success = _unmap (&mapinfo, py_mapinfo);
+  if (PyBool_Check (success) && mapinfo) {
+    /* Extract GstBuffer from Gst.Buffer parameter */
+    memory = GST_MEMORY_CAST (pygobject_get (py_memory));
+
+    /* Unmap the buffer, using reconstructed GstMapInfo */
+    gst_memory_unmap (memory, mapinfo);
+    g_free (mapinfo);
+  }
+
+  return success;
+}
+
+static PyObject *
+_gst_buffer_override_map_range (PyObject * self, PyObject * args)
+{
+  PyTypeObject *gst_buffer_type;
+  PyObject *py_buffer, *py_mapinfo, *success;
+  int flags, range;
+  unsigned int idx;
+  GstBuffer *buffer;
+  GstMapInfo *mapinfo;
+  _Bool ok;
+
+  /* Look up Gst.Buffer, Gst.MapInfo, idx, range, and Gst.MapFlags parameters */
+  gst_buffer_type = pygobject_lookup_class (_gst_buffer_type);
+  if (!PyArg_ParseTuple (args, "O!OIii", gst_buffer_type, &py_buffer,
+          &py_mapinfo, &idx, &range, &flags))
+    goto err;
+
+  /* Since Python does only support r/o or r/w it has to be changed to either */
+  flags = (flags & GST_MAP_WRITE) ? GST_MAP_READWRITE : GST_MAP_READ;
+
+  /* Extract GstBuffer from Gst.Buffer parameter */
+  buffer = GST_BUFFER (pygobject_get (py_buffer));
+
+  /* Map the buffer, fill out GstMapInfo */
+  mapinfo = g_new0 (GstMapInfo, 1);
+  ok = gst_buffer_map_range (buffer, idx, range, mapinfo, flags);
+  if (!ok) {
+    g_free (mapinfo);
+    goto err;
+  }
+
+  success = _remap (mapinfo, py_mapinfo);
+  if (!success) {
+    gst_buffer_unmap (buffer, mapinfo);
+    g_free (mapinfo);
+  }
+  return success;
+
+err:
+  Py_INCREF (Py_False);
+  return Py_False;
+}
+
+static PyObject *
+_gst_buffer_override_map (PyObject * self, PyObject * args)
+{
+  PyTypeObject *gst_buffer_type;
+  PyObject *py_buffer, *py_mapinfo, *success;
+  int flags;
+  GstBuffer *buffer;
+  GstMapInfo *mapinfo;
+  _Bool ok;
+
+  /* Look up Gst.Buffer, Gst.MapInfo, and Gst.MapFlags parameters */
+  gst_buffer_type = pygobject_lookup_class (_gst_buffer_type);
+  if (!PyArg_ParseTuple (args, "O!Oi", gst_buffer_type, &py_buffer, &py_mapinfo,
+          &flags)) {
+    PyErr_BadArgument ();
+    return NULL;
+  }
+
+  /* Since Python does only support r/o or r/w it has to be changed to either */
+  flags = (flags & GST_MAP_WRITE) ? GST_MAP_READWRITE : GST_MAP_READ;
+
+  /* Extract GstBuffer from Gst.Buffer parameter */
+  buffer = GST_BUFFER (pygobject_get (py_buffer));
+
+  /* Map the buffer, fill out GstMapInfo */
+  mapinfo = g_new0 (GstMapInfo, 1);
+  ok = gst_buffer_map (buffer, mapinfo, flags);
+  if (!ok) {
+    g_free (mapinfo);
+    goto err;
+  }
+
+  success = _remap (mapinfo, py_mapinfo);
+  if (!success) {
+    gst_buffer_unmap (buffer, mapinfo);
+    g_free (mapinfo);
+  }
+  return success;
+
+err:
+  Py_INCREF (Py_False);
+  return Py_False;
+}
+
+static PyObject *
+_gst_buffer_override_unmap (PyObject * self, PyObject * args)
+{
+  PyTypeObject *gst_buf_type;
+  PyObject *py_buffer, *py_mapinfo, *success;
+  GstBuffer *buffer;
+  GstMapInfo *mapinfo = NULL;
+
+  /* Look up Gst.Buffer and Gst.Mapinfo parameters */
+  gst_buf_type = pygobject_lookup_class (_gst_buffer_type);
+  if (!PyArg_ParseTuple (args, "O!O", gst_buf_type, &py_buffer, &py_mapinfo)) {
+    PyErr_BadArgument ();
+    return NULL;
+  }
+
+  success = _unmap (&mapinfo, py_mapinfo);
+  if (PyBool_Check (success) && mapinfo) {
+    /* Extract GstBuffer from Gst.Buffer parameter */
+    buffer = GST_BUFFER (pygobject_get (py_buffer));
+
+    /* Unmap the buffer, using reconstructed GstMapInfo */
+    gst_buffer_unmap (buffer, mapinfo);
+    g_free (mapinfo);
+  }
+
+  return success;
+}
+
 static PyMethodDef _gi_gst_functions[] = {
   {"trace", (PyCFunction) _wrap_gst_trace, METH_VARARGS,
       NULL},
@@ -756,6 +1032,19 @@ static PyMethodDef _gi_gst_functions[] = {
       NULL},
   {"memdump", (PyCFunction) _wrap_gst_memdump, METH_VARARGS,
       NULL},
+  {"buffer_override_map_range", (PyCFunction) _gst_buffer_override_map_range,
+        METH_VARARGS,
+      NULL},
+  {"buffer_override_map", (PyCFunction) _gst_buffer_override_map, METH_VARARGS,
+      NULL},
+  {"buffer_override_unmap", (PyCFunction) _gst_buffer_override_unmap,
+        METH_VARARGS,
+      NULL},
+  {"memory_override_map", (PyCFunction) _gst_memory_override_map, METH_VARARGS,
+      NULL},
+  {"memory_override_unmap", (PyCFunction) _gst_memory_override_unmap,
+        METH_VARARGS,
+      NULL},
   {NULL, NULL, 0, NULL}
 };
 
