@@ -32,7 +32,8 @@ import gc
 import unittest
 import gi.overrides
 
-
+import gi
+gi.require_version("Gst", "1.0")
 from gi.repository import Gst
 
 
